

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>
#include <math.h>
#include "words.h"
#include "md5_crypt.h"

int pool_bits_left=0;
unsigned rand_pool=0;

#define min(A, B)  ((A) < (B) ? (A) : (B))

unsigned getrandbits(int bits);
void print_binary(unsigned in);
void refresh_pool();
char *get_char_passwd(int length);
char *get_word_passwd(int length);
char *generate_md5_pw(const char *passwd);

int main(int argc, char *argv[])
{
	char c;
	char *passwd, *md5_pw;

	int length = 0;
	int bit_count = 70;
	int word_based = 1;
	int should_generate = 0;

	while((c=getopt(argc, argv, "l:wcgb:")) != EOF) {
		switch(c) {
			case 'l':
				length = atoi(optarg);
				break;
			case 'w':
				word_based = 1;
				break;
			case 'c':
				word_based = 0;
				break;
			case 'b':
				bit_count = atoi(optarg);
				break;
			case 'g':
				should_generate = 1;
				break;
			default:
				fprintf(stderr, "rpasswd version 0.1\n");
				fprintf(stderr, "Useage %s [-l length] [-b bits] [-g] [-c] [-w]\n", argv[0]);
				fprintf(stderr, "Where:\n");
				fprintf(stderr, "-l\tLength of password (in characters or words)\n");
				fprintf(stderr, "-b\tEffective password entropy\n");
				fprintf(stderr, "-g\tGenerate a MD5 password hash too\n");
				fprintf(stderr, "-c\tGenerate a character based password\n");
				fprintf(stderr, "-w\tGenerate a word based password\n");
				exit(-1);
		}
	}

	if(!length) {
		if(word_based) {
			length = bit_count / 11;
			if(bit_count % 11)
				length++;
		} else
			length = ceil(((float)bit_count)/6.5546);
	}

	if(word_based)
		passwd = get_word_passwd(length);
	else
		passwd = get_char_passwd(length);

	printf("%s\n", passwd);

	if(should_generate) {
		md5_pw = generate_md5_pw(passwd);
		printf("%s\n", md5_pw);
	}

	return(0);
}

char *get_char_passwd(int length)
{
	char *passwd;
	int current_length=0;
	unsigned char c;

	if((passwd = (char *) malloc(length+1)) == NULL) {
		perror("Error allocating memory");
		exit(-1);
	}

	while(current_length < length) {
		c = getrandbits(7);
		if(c < 94) {
			passwd[current_length] = c + 33;
			current_length++;
		}
	}

	passwd[length] = '\0';
	return passwd;
}

char *get_word_passwd(int length)
{
	int output_length;
	int current_length;
	char *passwd;

	output_length = 6 * length;

	if((passwd = (char *) malloc(output_length+1)) == NULL) {
		perror("Error allocating memory");
		exit(-1);
	}

	strcpy(passwd, words[getrandbits(11)]);
	for(current_length=1; current_length < length; current_length++) {
		strcat(passwd, " ");
		strcat(passwd, words[getrandbits(11)]);
	}

	return passwd;
}

char *generate_md5_pw(const char *passwd)
{

	char *cryptpw;
	char salt[9];
	unsigned int salt_value;

	salt_value = getrandbits(24);
	to64(salt, salt_value, 4);
	salt_value = getrandbits(24);
	to64(salt+4, salt_value, 4);

	salt[8] = '\0';

	cryptpw = crypt_md5(passwd, salt);

	return cryptpw;
}


/*
This gets a certain number of random bits from some random source
(could be an OS provided thing, some interface to a physicly-random
source, etc).  The returned value is an unsigned int that is in the
range [0, 2^bits)... That is, between 0 (inclusive) and 2 raised to
the number of bits (exclusive)
*/

void print_binary(unsigned in)
{
	int cntr;
	for(cntr=0; cntr < 32; cntr++) {
		if(!(cntr % 4))
			printf(" ");
		printf("%c", ((in & (1 << (31 - cntr)))? '1' : '0'));
	}
}
		
unsigned getrandbits(int requested_bits)
{
	int current_bits=0;
	int current_round_bits=0;
	unsigned rand_bits=0;

	assert(requested_bits <= (sizeof(unsigned) * 8));

	while(current_bits < requested_bits) {
		if(!pool_bits_left)
			refresh_pool();

		/*grab some from the pool*/
		current_round_bits = min(pool_bits_left, requested_bits - current_bits);
		rand_bits <<= current_round_bits;
		current_bits += current_round_bits;
		rand_bits |= (rand_pool & ~(~0 << current_round_bits));
		rand_pool >>= current_round_bits;
		pool_bits_left -= current_round_bits;
	}

	return rand_bits;
}
		
void refresh_pool()
{
	int fd;
	if((fd = open("/dev/random", O_RDONLY)) < 0) {
		perror("Error opening /dev/random");
		exit(-1);
	}
	
	if (read(fd, &rand_pool, sizeof(unsigned)) < sizeof(unsigned)) {
		perror("Error reading from /dev/random");
		exit(-1);
	}

	pool_bits_left = sizeof(unsigned) * 8;

	close(fd);
}
